/*
 * ============================================================================
 *  Title:    MAIN
 *  Author:   J. Zbiciak
 *  $Id: jzintv.c,v 1.31 1999/09/17 13:10:00 im14u2c Exp $
 * ============================================================================
 *  Main Simulator Driver File
 * ============================================================================
 *  This doesn't do much yet.  :-)
 * ============================================================================
 */

static const char rcs_id[]="$Id: jzintv.c,v 1.31 1999/09/17 13:10:00 im14u2c Exp $";

#include "config.h"

#ifndef macintosh
#include <SDL/SDL.h>            /* Ack!  Needed for hacks to main()     */
#endif

#include <signal.h>
#include "macros.h"
#include "plat/plat.h"
#include "periph/periph.h"
#include "cp1600/cp1600.h"
#include "mem/mem.h"
#include "pads/pads.h"
#include "gfx/gfx.h"
#include "snd/snd.h"
#include "stic/stic.h"
#include "ay8910/ay8910.h"
#include "debug/debug.h"
#include "speed/speed.h"
#include "event/event.h"
#include "file/file.h"

#ifdef macintosh
# include "console.h"

/* Prototypes required for Macintosh */
double elapsed(int);
#endif

periph_bus_p    intv;           /* The Intellivision Console            */
cp1600_t        cp1610;         /* CPU                                  */
stic_t          stic;           /* STIC                 0000-0032   16  */
/*ivoice_t      ivoice;*/       /* Intellivoice         0080-00FF    8  */
mem_t           scr_ram;        /* Scratch RAM          0100-01EF    8  */
mem_t           fake_psg;       /* Fake Sound Chip      01F0-01FF    8  */
ay8910_t        ay8914;         /* AY8914 Sound Chip    01F0-01FD    8  */
pad_t           pad0,pad1;      /* Controller pad 0, 1  01FE-01FF    8  */
mem_t           sys_ram;        /* System RAM           0200-035F   16  */
mem_t           exec_rom;       /* EXEC ROM             1000-1FFF   10  */
mem_t           grom;           /* Graphics ROM         3000-37FF    8  */
mem_t           gram;           /* Graphics RAM         3800-39FF    8  */
mem_t           game0;          /* Game Cart            5000-6FFF   10  */
mem_t           game1;          /* Game Cart            D000-DFFF   10  */
mem_t           game2;          /* Game Cart            F000-FFFF   10  */
debug_t         debug;          /* Debugger             0000-FFFF   16  */

gfx_t           gfx;            /* Graphics display                     */
snd_t           snd;            /* Sound mixer                          */
speed_t         speed;          /* Speed controller                     */
event_t         event;          /* Event handler.                       */

uint_16 game_img[32768];
uint_16 exec_img[4096];
uint_16 grom_img[2048];
uint_16 ay8914_img[16];

volatile int please_die = 0;
volatile int reset = 0;

static int read_rom(char * filename, int size, uint_16 * image);
static uint_32 checksum(uint_16 *mem, int len) UNUSED;
static void usage(void);
void dump_state(void);

/*
 * ============================================================================
 *  RELEASE      -- Program name / release
 * ============================================================================
 */
static char * release(void)
{
    char *s2;
    const char *Name = "$Name: release_010a $", *s1;
    static char name_buf[64];
    static int named = 0;

    if (named) return name_buf;

    s1 = Name;

    while (*s1 && *s1 != ':') s1++;
    if (*s1 == ':') s1++;
    while (*s1 && isspace(*s1) && *s1 != '$') s1++;

    if (!*s1 || *s1 == '$')
    {
        strncpy(name_buf, "[unreleased]", sizeof(name_buf)-1);
        name_buf[sizeof(name_buf)-1] = 0;
    } else
    {
        s2 = name_buf;
        while (s2 < (name_buf + sizeof(name_buf) - 1) && *s1 && *s1!='$')
            *s2++ = *s1++;
        *s2++ = 0;
    }
    
    named = 1;

    return name_buf;
}

/*
 * ============================================================================
 *  CART_NAME    -- Look for a game name in a cartridge image.
 * ============================================================================
 */
static char * cart_name(uint_16 *game, uint_16 addr, uint_16 size)
{
    static char name_buf[64];
    char * s1;
    uint_16 *s2;
    uint_16 ofs;
    int len = 0;

    s2  = game + 0x000A;
    ofs = (s2[1] << 8) + (s2[0]) - addr + 1;


    if (ofs > size) { strcpy(name_buf, "[unknown]"); return name_buf; } 

    s2 = game + ofs;
    s1 = name_buf;

    while ((*s2 & 0xFF) && len++ < 60)
        *s1++ = *s2++;

    return name_buf;
}

/*
 * ============================================================================
 *  GRACEFUL_DEATH -- Die gracefully.
 * ============================================================================
 */
static void graceful_death(int x)
{
    if (please_die < 2)
    {
        if (please_die) fprintf(stderr, "\nOUCH!\n");
        fprintf(stderr, 
                "\nRequesting exit:  Received signal %d.\n"
                "(You may need to press enter if you're at a prompt.)\n", x);
    } else
    {
        fprintf(stderr, "\nReceived 3 signals:  Aborting on signal %d.\n", x);
        exit(1);
    }
    please_die++;
}

/*
 * ============================================================================
 *  ELAPSED      -- Returns amount of time that's elapsed since the program
 *                  started, in CP1610 clock cycles (895kHz)
 * ============================================================================
 */
double elapsed(int restart)
{
    static struct timeval start;
    static int init = 0;
    struct timeval now;
    uint_32 usec, sec;

    if (!init || restart) 
    {
        gettimeofday(&start, NULL);
        init = 1;
    }

    gettimeofday(&now, NULL);

    if (now.tv_usec < start.tv_usec)
    {
        now.tv_usec += 1000000;
        now.tv_sec--;
    }

    usec = now.tv_usec - start.tv_usec;
    sec  = now.tv_sec  - start.tv_sec;

    return (sec + usec/1000000.) * 894886.;
}

/*
 * ============================================================================
 *  In the beginning, there was a main....
 * ============================================================================
 */
#ifdef WIN32
main(int argc, char *argv[]) 
#else
int main(int argc, char *argv[]) 
#endif
{ 
    int iter = 0, addr;
    double cycles = 0, rate, irate, then, now, icyc;
    int debugging = 0;
    int snd_rate = 44100;
    int snd_window = -1;
    int arg = 0, do_speed = 1;
    char *fn_exec = "exec.bin";
    char *fn_grom = "grom.bin";
    char *fn_game = "game.bin", *na;
    char **fn = &na;
    char title[128];
    char *raw_file = NULL;
    uint_32 s_cnt = 0;

    uint_32 hcif_io_base = 0;
    int     full_screen  = 0;           /* 0 = window,  1 = full screen */
    int     res          = 0;           /* 0 = 320x200, 1 = 640x480     */
    int     res_x[2]     = { 320, 640 };
    int     res_y[2]     = { 200, 480 };
    int     res_d[2]     = { 8,   8   };

    /* -------------------------------------------------------------------- */
    /*  Platform-specific initialization.                                   */
    /* -------------------------------------------------------------------- */
    if (plat_init())
    {
        fprintf(stderr, "Error initializing.\n");
        exit(1);
    }

    /* -------------------------------------------------------------------- */
    /*  Parse arguments.  The commandline is of the following form:         */
    /*                                                                      */
    /*      jzintv [-e exec.bin] [-g grom.bin] [-d] [game.bin]              */
    /*                                                                      */
    /*      -e        specifies a different exec.bin file to load.          */
    /*      -g        specifies a different grom.bin file to load.          */
    /*      -d        enables debugging mode.                               */
    /*      -s        toggle the sprite-hack                                */
    /*      -p        toggle the pad-hack                                   */
    /*      -r        toggle rate control (speed_t)                         */
    /*      -a#       specifies audio sampling rate (0 == off)              */
    /*      -f file   specifies a file to dump raw audio samples to.        */
    /*      -w#       changes the sound-averaging window (AY8910 only)      */
    /*      -i###     set hand-controller interface port #.                 */
    /*      -x        toggle full-screen mode                               */
    /*      -r#       Set display resolution: 0 = 320x200, 1 = 640x480      */
    /*                                                                      */
    /*  TODO: Do this right w/ getopt() or the like.                        */
    /* -------------------------------------------------------------------- */

#ifdef macintosh
    argc = ccommand( &argv );
#endif

    for (arg = 1; arg < argc; arg++)
    {
        if (argv[arg][0] != '-')
        {
            fn_game = strdup(argv[arg]);
            continue;
        }
        fn = &fn_grom;
        switch (argv[arg][1])
        {
            case 'e' : case 'E' : fn = &fn_exec; /* fall through */
            case 'g' : case 'G' :
            {
                if (++arg >= argc)
                {
                    fprintf(stderr, "Error: -%c requires an argument\n",
                            argv[arg][1]);
                    exit(1);
                }
                *fn = strdup(argv[arg]);
                break;
            }
            case 'f': case 'F':
            {
                if (++arg >= argc)
                {
                    fprintf(stderr, "Error: -%c requires an argument\n",
                            argv[arg][1]);
                    exit(1);
                }
                raw_file = strdup(argv[arg]);
                break;
            }
            case 'x' : case 'X' :
            {
                full_screen = !full_screen;
                break;
            }
            case 'z' : case 'Z' :
            {
                res = atoi(argv[arg]+2);
                if (res >= (int)(sizeof(res_x) / sizeof(res_x[0])) || res < 0)
                {
                    printf("Invalid resolution %d\n", res);
                    exit(1);
                }
                break;
            }
            case 'd' : case 'D' :
            {
                debugging = 1;
                do_speed = 0;
                break;
            }
            case 's' : case 'S' :
            {
                extern int sprite_hack;

                sprite_hack = !sprite_hack;
                break;
            }
            case 'a' : case 'A' :
            {
                snd_rate = atoi(argv[arg]+2);
                break;
            }
            case 'w' : case 'W' :
            {
                snd_window = atoi(argv[arg]+2);
                if (snd_window < 1) 
                    snd_window = -1;
                break;
            }
            case 'r' : case 'R' :
            {
                do_speed = !do_speed;
                break;
            }
#ifdef DIRECT_HCIF
            case 'i' : case 'I' : 
            {
                if (argv[arg][2] != 0)
                {
                    if (argv[arg][2] == '0' && tolower(argv[arg][3]) == 'x')
                        sscanf(argv[arg]+4, "%x", &hcif_io_base);
                    else
                        sscanf(argv[arg]+2, "%x", &hcif_io_base);
                } else
                {
                    hcif_io_base = 0x378;
                }
                break;
            }
#endif
            case '?' : case 'h' : case 'H' :
            {
                usage();
                exit(0);
            }
            default:
            {
                fprintf(stderr, "Unrecognized option -%c\n", argv[arg][1]);
                usage();
                exit(1);
            }
        }
    }

    /* -------------------------------------------------------------------- */
    /*  Load up the ROM images.                                             */
    /* -------------------------------------------------------------------- */

    read_rom(fn_exec, 4096, exec_img);
    read_rom(fn_grom, 2048, grom_img);
    read_rom(fn_game,32768, game_img);

    /* -------------------------------------------------------------------- */
    /*  Check to see if the grom image we read was 8-bit or 16-bit.  Fix    */
    /*  it if necessary.  This test should work because even 16-bit ROM     */
    /*  images should store a 0 in the upper byte for all of the words.     */
    /* -------------------------------------------------------------------- */
    for (addr = 0; addr < 2048; addr++)
        if (grom_img[addr] & 0xFF00) break;
    
    if (addr != 2048)
    {
        for (addr = 1023; addr >= 0; addr--)
        {
            grom_img[addr*2 + 1] =  grom_img[addr]       & 0xFF;
            grom_img[addr*2 + 0] = (grom_img[addr] >> 8) & 0xFF;
        }
    }


    for (addr = 0; addr < 16; addr++)
        ay8914_img[addr] = 0xFF;

    /* -------------------------------------------------------------------- */
    /*  Create a new peripheral bus for the Intellivision main console.     */
    /* -------------------------------------------------------------------- */
    intv = periph_new(16, 16, 4);
    strncpy(intv->periph.name, "Intellivision", 16);


    /* -------------------------------------------------------------------- */
    /*  Initialize the peripherals.                                         */
    /* -------------------------------------------------------------------- */
    if (gfx_init(&gfx, res_x[res], res_y[res], res_d[res], full_screen))
    {
        printf("Failed to initialize graphics\n");
        exit(1);
    }

    if (snd_rate && snd_init(&snd, snd_rate, raw_file))
    {
        printf("Failed to initialize sound\n");
        exit(1);
    }

    if (cp1600_init(&cp1610, 0x1000, 0x1004))
    {
        printf("Failed to initialize CP1610\n");
        exit(1);
    }

    if (mem_make_ram(&scr_ram,  8, 0x0100, 8) ||
        mem_make_ram(&sys_ram, 16, 0x0200, 9) ||
        mem_make_ram(&gram,     8, 0x3800, 9))
    {
        printf("Failed to initialize RAMs\n");
        exit(1);
    }

#if 1
    if (mem_make_rom(&exec_rom,10, 0x1000, 12, exec_img) ||
        mem_make_rom(&grom,     8, 0x3000, 11, grom_img) ||
        mem_make_rom(&game0,   10, 0x5000, 13, game_img         ) ||
        mem_make_rom(&game1,   10, 0x9000, 13, game_img + 8192  ) ||
        mem_make_rom(&game2,   10, 0xF000, 12, game_img + 4096*3) ||
       (!snd_rate && 
        mem_make_rom(&fake_psg, 8, 0x01F0, 4,  ay8914_img)))
    {
        printf("Failed to initialize ROMs\n");
        exit(1);
    }
#else
    if (mem_make_rom(&exec_rom,10, 0x1000, 12, exec_img) ||
        mem_make_rom(&grom,     8, 0x3000, 11, grom_img) ||
        mem_make_rom(&game0,   10, 0x6000, 12, game_img         ) ||
        mem_make_rom(&game1,   10, 0x7000, 12, game_img + 4096  ) ||
       (!snd_rate && 
        mem_make_rom(&fake_psg, 8, 0x01F0, 4,  ay8914_img)))
    {
        printf("Failed to initialize ROMs\n");
        exit(1);
    }
#endif


    /* Temporary hack so ^C, kill work */
    signal( 2, graceful_death);
    signal(15, graceful_death);

    if (stic_init(&stic, &grom, &gram, &cp1610.irq, &gfx))
    {
        printf("Failed to initialize STIC\n");
        exit(1);
    }

    if (snd_rate && ay8910_init(&ay8914, 0x01F0, &snd, snd_rate, snd_window))
    {
        printf("Failed to initialize AY8910\n");
        exit(1);
    }

    if (debugging && debug_init(&debug, &cp1610))
    {
        printf("Failed to initialize debugger\n");
        exit(1);
    }
    
    if (pad_init(&pad0, 0x1F0, hcif_io_base))
    {
        printf("Failed to initialize game pads\n");
        exit(1);
    }

    if (do_speed && speed_init(&speed, &gfx))
    {
        printf("Failed to initialize rate control.\n");
        exit(1);
    }

    if (event_init(&event))
    {
        printf("Failed to initialize event subsystem.\n");
        exit(1);
    } else
    {
        extern void cfg_default(event_t *);

        cfg_default(&event);
    }
    
    /* -------------------------------------------------------------------- */
    /*  Register the CPU, system RAMs and ROMs.                             */
    /* -------------------------------------------------------------------- */
    periph_register(intv, (periph_p)&cp1610,  0x0000, 0x0000, "CP-1610");
    if (snd_rate)
    {
        periph_register(intv, (periph_p)&ay8914,  0x01F0, 0x01FF, "PSG AY8914");
        periph_register(intv, (periph_p)&snd,     0x0000, 0x0000, "Sound");   
    } else
        periph_register(intv, (periph_p)&fake_psg,0x01F0, 0x01FF, "Fake PSG");

    periph_register(intv, (periph_p)&gfx,     0x0000, 0x0000, "Graphics");
    periph_register(intv, (periph_p)&scr_ram, 0x0100, 0x01EF, "Scratch RAM");
    periph_register(intv, (periph_p)&sys_ram, 0x0200, 0x035F, "System RAM");
    periph_register(intv, (periph_p)&exec_rom,0x1000, 0x1FFF, "EXEC");
    periph_register(intv, (periph_p)&grom,    0x3000, 0x37FF, "GROM");
    periph_register(intv, (periph_p)&gram,    0x3800, 0x3FFF, "GRAM");
#if 1
    periph_register(intv, (periph_p)&game0,   0x5000, 0x6FFF, "Game ROM 0");
    periph_register(intv, (periph_p)&game1,   0x9000, 0xDFFF, "Game ROM 1");
    periph_register(intv, (periph_p)&game2,   0xF000, 0xFFFF, "Game ROM 2");
#else
    periph_register(intv, (periph_p)&game0,   0x6000, 0x6FFF, "Game ROM 0");
    periph_register(intv, (periph_p)&game1,   0x7000, 0x7FFF, "Game ROM 1");
#endif
    periph_register(intv, (periph_p)&pad0,    0x01F0, 0x01FF, "Pad 0");
    /*periph_register(intv, (periph_p)&pad1,    0x01FE, 0x01FE, "Pad 1");*/
    periph_register(intv, &stic.stic_cr,      0x0000, 0x0032, "STIC");
    periph_register(intv, &stic.stic_cr,      0x4000, 0x4032, "STIC alias 1");
    periph_register(intv, &stic.stic_cr,      0x8000, 0x8032, "STIC alias 2");
    periph_register(intv, &stic.snoop_btab,   0x0200, 0x02EF, "STIC (BTAB)");
    periph_register(intv, &stic.snoop_gram,   0x3800, 0x3FFF, "STIC (GRAM)");
    periph_register(intv, (periph_p)&event,   0x0000, 0x0000, "Event");

    if (do_speed)
        periph_register(intv, (periph_p)&speed,   0x0000, 0x0000, "Rate Ctrl");

    /* -------------------------------------------------------------------- */
    /*  Mark the ROMs cacheable in the CPU.  Mark the RAMs as needing       */
    /*  bus-snoop support.                                                  */
    /* -------------------------------------------------------------------- */
    cp1600_cacheable(&cp1610, 0x0100, 0x01EF, 1);
    cp1600_cacheable(&cp1610, 0x0200, 0x035F, 1);
    cp1600_cacheable(&cp1610, 0x1000, 0x1FFF, 0);
    cp1600_cacheable(&cp1610, 0x3000, 0x37FF, 0);
    cp1600_cacheable(&cp1610, 0x3800, 0x3FFF, 1);
    cp1600_cacheable(&cp1610, 0x5000, 0x6FFF, 0);
    cp1600_cacheable(&cp1610, 0x9000, 0xDFFF, 0);
    cp1600_cacheable(&cp1610, 0xF000, 0xFFFF, 0);

    /* -------------------------------------------------------------------- */
    /*  Register the debugger.  This _must_ be done last.                   */
    /* -------------------------------------------------------------------- */
    if (debugging)
        periph_register(intv, (periph_p)&debug,   0x0000, 0xFFFF, "Debugger");

    /* -------------------------------------------------------------------- */
    /*  Set the window title.  If we recognize a standard Intellivision     */
    /*  ROM header at 0x5000, then also include the cartridge name.         */
    /* -------------------------------------------------------------------- */
    snprintf(title, 128, "jzintv %.32s : %.64s", 
             release(), cart_name(game_img, 0x5000, 0x1000));
    gfx_set_title(&gfx,title);

    /* -------------------------------------------------------------------- */
    /*  Run the simulator.                                                  */
    /* -------------------------------------------------------------------- */
    if (debugging)
        debug_tk((periph_p)&debug,1);

    printf("Starting jzIntv...\n");
    fflush(stdout);
    sleep(1);
restart:

    iter = 1;
    now = elapsed(1);
    while (now == elapsed(0))    /* Burn some time. */
        ;

    icyc   = 0;
    s_cnt  = 0;
    cycles = 0;
    speed_resync(&speed);

    while (please_die == 0)
    {
        cycles += periph_tick((periph_p)intv, 
                               debugging ? 1 : reset ? 1000 : 3579545/240);

        if (!debugging && !reset && (iter++&63) == 0)
        {
            then  = now;
            now   = elapsed(0);
            rate  = (cycles / now);
            irate = (cycles - icyc) / (now - then);
            icyc  = cycles;

            printf("Rate: [%6.2f%% %6.2f%%]  Drop Gfx:[%6.2f%% %6d] "
                   "Snd:[%6.2f%% %6d %6.3f]\r", 
                    rate * 100., irate * 100.,
                    100. * gfx.tot_dropped_frames / gfx.tot_frames, 
                    (int)gfx.tot_dropped_frames,
                    100. * snd.mixbuf.tot_drop / snd.tot_frame, 
                    (int)snd.mixbuf.tot_drop,
                    (double)snd.tot_dirty / snd.tot_frame);
            fflush(stdout);
        }

        if (reset)
        {
            cp1610.r[7] = 0x1000;
            gfx_vid_enable(&gfx, 0);
            s_cnt++;
        } else 
        {
            if (s_cnt > 1500) break;
            s_cnt = 0;
        }
    }

    s_cnt = 0x2A3A4A5A;
    arg = 0;
    while (please_die == 0)
    {
        int i, j;
        uint_8 p;

        if (reset) arg = 1;
        if (!reset && arg) 
        {
            p = stic.cr.raw[0x2C] & 15;
            for (i = 0; i < 160 * 200; i++)
                stic.disp[i] = p;

            goto restart;
        }
        
        for (i = 0; i < 160 * 200; i++)
        {
            for (j = 0; j < 16; j++)
                s_cnt = (s_cnt << 1) | (1 & ((s_cnt >> 28) ^ (s_cnt >> 30)));


            p = (s_cnt & 0xF) + 16;
            if (p == 16) p = 0;

            stic.disp[i] = p;
        }

        gfx.dirty = 1;
        gfx.periph.tick  ((periph_p)&gfx,   0);
        event.periph.tick((periph_p)&event, 0);
    }

    if (please_die)
        printf("Exited on user request.\n");

    return 0;
}


#include "cp1600/op_decode.h"
#include "cp1600/op_exec.h"

/*
 * ============================================================================
 *  DUMP_STATE   -- Called by op_exec:fn_HLT.  Dumps state of universe.
 * ============================================================================
 */

void dump_state(void)
{
    FILE *f;
    int addr, data, i, j;

    f = fopen("dump.mem","w");

    debug.show_rd = 0;
    debug.show_wr = 0;
    for (addr = 0; addr <= 0xFFFF; addr++)
    {   
        data = periph_read((periph_p)intv, (periph_p)intv, addr, 0);
        fputc((data >> 8) & 0xFF, f);
        fputc((data     ) & 0xFF, f);
    }

    fclose(f);

    f = fopen("dump.cpu", "w");

    fprintf(f, "CP-1600 State Dump\n");
    fprintf(f, "Tot Cycles:   %lld\n", cp1610.tot_cycle);
    fprintf(f, "Tot Instrs:   %lld\n", cp1610.tot_instr);
    fprintf(f, "Tot Cache:    %d\n", cp1610.tot_cache);
    fprintf(f, "Tot NonCache: %d\n", cp1610.tot_noncache);
    fprintf(f, "Registers:    %.4x %.4x %.4x %.4x %.4x %.4x %.4x %.4x\n",
            cp1610.r[0], cp1610.r[1], cp1610.r[2], cp1610.r[3],
            cp1610.r[4], cp1610.r[5], cp1610.r[6], cp1610.r[7]);
    fprintf(f, "Flags:        S:%d C:%d O:%d Z:%d I:%d D:%d intr:%d irq:%d\n",
            cp1610.S, cp1610.C, cp1610.O, cp1610.Z, cp1610.I, cp1610.D,
            cp1610.intr, cp1610.irq);

    fprintf(f, "Cacheability Map:\n");

    for (i = 0; i < 1 << (CP1600_MEMSIZE-CP1600_DECODE_PAGE - 5); i++)
    {
        addr = (i << (CP1600_DECODE_PAGE + 5));

        fprintf(f, "   %.4x-%.4x:", addr, addr+(32<<CP1600_DECODE_PAGE)-1);
        for (j = 0; j < 32; j++)
        {
            fprintf(f, " %d", 1 & (cp1610.cacheable[i] >> j));
        }
        fprintf(f, "\n");
    }


    fprintf(f,"Decoded Instruction Map:\n");

    for (i = 0; i < 1 << (CP1600_MEMSIZE - 6); i++)
    {
        addr = i << 6;

        fprintf(f, "   %.4x-%.4x:", addr, addr + 63);
        for (j = 0; j < 64; j++)
        {
            fprintf(f, "%c", 
                (void*)cp1610.execute[addr + j] == (void*)fn_decode_1st ? '-' :
                (void*)cp1610.execute[addr + j] == (void*)fn_decode     ? 'N' :
                (void*)cp1610.execute[addr + j] == (void*)fn_invalid    ? '!' :
                                                                          'C');
        }
        fprintf(f, "\n");
    }

    fclose(f);
}


/*
 * ============================================================================
 *  READ_ROM     -- Reads a ROM image file, adjusting it for endian if 
 *                  necessary
 * ============================================================================
 */

int read_rom(char * filename, int size, uint_16 * image)
{
    FILE *f;
    int tot = 0, block, addr;

#ifdef macintosh
# pragma unused( addr )
#endif

    memset(image, 0xFF, size * 2);

    f = fopen(filename, "rb");

    if (!f)
    {
        fprintf(stderr, "Could not open '%s' for reading.\n", filename);
        perror("fopen()"); exit(1);
    }

    while (tot < size)
    {
        block = size - tot > 128 ? 128 : size - tot;
        block = fread(image + tot, 2, block, f);
        tot += block;

        if (block < 0)
        {
            fprintf(stderr, "Error while reading '%s'\n", filename);
            perror("fread()");  exit(1);
        }
        if (block == 0)
            break;
    }
    fclose(f);

#ifdef _LITTLE_ENDIAN
    for (addr = 0; addr < size; addr++)
    {
        image[addr] = (0x00FF & (image[addr] >> 8)) | 
                      (0xFF00 & (image[addr] << 8));
    }
#endif

    printf("Read %d words from '%s'.\n", tot, filename);
    return tot;
}

/*
 * ============================================================================
 *  CHECKSUM     -- Returns a simple checksum on a region of memory.
 *                  I was using this to try to find some bugs, assuming
 *                  memory corruption was happening.
 * ============================================================================
 */

static uint_32 checksum(uint_16 *mem, int len)
{
    uint_32 cksum = 0x2A3A4A5A;

    while (len-->0)
    {
        cksum ^= *mem;
        cksum  = (cksum >> 1) | (cksum << 31);
        cksum += *mem++;
    }

    return cksum;
}

/*
 * ============================================================================
 *  USAGE        -- Prints a help message.  Whee.
 * ============================================================================
 */

static void usage(void)
{
    printf
    (
        "jzIntv %s\n\n"
        "usage: jzintv [-d] [-v] [-e exec_img] [-g grom_img] [game_img]\n"
        "       jzintv [-?|-h]\n"
        "\n"
        "       -d      Enables debugger\n"
        "       -s      Toggles sprite-boxing\n"
        "       -p      Toggles pad-input sequencer\n"
        "       -e      Specifies alternate EXEC image\n"
        "       -g      Specifies alternate GROM image\n"
        "       -a###   Specifies audio sample rate.  -a0 disables sound\n"
        "       -w###   Specifies window size for AY8910 sliding window avg\n"
        "       -f raw  Dumps audio to raw sample file 'raw'\n"
        "       -? -h   This help info.\n"
#ifdef DIRECT_HCIF
        "       -i###   Enable Direct Hand-Controller Interface support on\n"
        "               port number ###.  -h alone defaults to 0x378.\n"
#endif
        "       -x      Attempt full-screen mode.\n"
        "       -z#     Set display resolution (0 = 320x200, 1 = 640x480\n"
        "\n"
        "       By default, jzintv looks for 'exec.bin', 'grom.bin' and\n"
        "       'game.bin' in the current directory.  These are overridden\n"
        "       by the command-line options given above.\n\n",
        release()
    );
    return;
}

/* ======================================================================== */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1998-1999, Joseph Zbiciak                  */
/* ======================================================================== */

